'use strict';
const CONST = require("../../constants.js");
const LOG   = require("../../core/Logger.js");

const BagBasedHandlerBase = require("./BagBasedHandlerBase.js");
const HandlerException = require("./HandlerException.js");

const BagUpdateDataModel = require("../dal/BagUpdateDataModel.js");

class DataHandler extends BagBasedHandlerBase {
	static getRoute() {
		return "/:bagId/data";
	}

	async _processGET() {
		let lastData = this._bag.getLastData();
		if (!lastData) return this.respondNotFound("No last data found");
		return this.respondOk({
			data : lastData.getData()
		});
	}

	async _processPUT() {
		return this.respondMethodNotAllowed("PUT");
	}

	async _processPOST() {
		return this._validateData(this._body)
			.then(r => this._storeData(r))
			.then(r => this.respondOk())
			.catch(e => this._handleError(e));
	}

	async _processDELETE() {
		return this.respondMethodNotAllowed("DELETE");
	}

	_validateData(data) {
		LOG.info("Data received", data);
		return Promise.resolve(data);
	}

	_updateBagData(record) {
		if (!record) return;
		if (!this._bag.getLastData() || this._bag.getLastData().getTimeStamp() < record.getTimeStamp()) {
			this._bag.setLastData(record);
			return this._saveBagData();
		}
	}

	_sendNotification(notificationData, title, message) {
		try {
			return this._firebase.messaging().sendToDevice(notificationData.getKey(), {
				notification: {
					title : title,
					body  : message
				},
				data : {
					bagId : this._bag.getKey()
				}
			}).catch(e => LOG.error("Failed to send message, '" + title + "'", e));
		} catch (e) {
			LOG.error("Failed to send message, '" + title + "'", e);
		}
	}

	_sendNotifications() {
		let alertSettings = this._bag.getAlertData();
		let notiData = this._bag.getNotificationData();
		let lastData = this._bag.getLastData();
		if (!alertSettings || !notiData || !lastData) return;

		let safePlaces = this._bag.getSafePlaces();
		let safePlace  = false;

		for (let i in safePlaces) {
			let place = safePlaces[i];
			let lon   = lastData.getLongitude();
			let lat   = lastData.getLatitude();
			if (Math.square(lon - place.lon) + Math.square(lat - place.lat) <= place.rr) {
				safePlace = true;
			}
		}

		let ret = [];
		let now = Date.now();
		if (alertSettings.alertIfLowBattery() && lastData.getBattery() < 25) {
			let lastNot = this._bag.getLastBatteryNotificationTs();
			if (now - lastNot > CONST.NOTIFICATION_PERIOD) {
				LOG.info("Sending alert for batter low");
				ret.push(this._sendNotification(notiData, "SMB Battery low", "The smartbag's battery is below 25%!"));
				this._bag.setLastBatteryNotificationTs(now);
			} else {
				LOG.info("skipping alert for battery low, notification have been already sent", lastNot, now);
			}
		}

		if (alertSettings.alertIfNoPhone() && !safePlace && !lastData.isPhoneActive()) {
			let lastNot = this._bag.getLastPositionNotificationTs();
			if (now - lastNot > CONST.NOTIFICATION_PERIOD) {
				LOG.info("Sending alert for phone inactive");
				ret.push(this._sendNotification(notiData, "SMB disconnect from phone", "The smartbag has lost connection with the associated phone!"));
				this._bag.setLastPositionNotificationTs(now);
			} else {
				LOG.info("skipping alert for phone inactive, notification have been already sent", lastNot, now);
			}
		}

		if (alertSettings.alertIfMissingGadget() && !safePlace && lastData.getMissingGadgets() > 0) {
			let lastNot = this._bag.getLastGadgetNotificationTs();
			if (now - lastNot > CONST.NOTIFICATION_PERIOD) {
				LOG.info("Sending alert for missing gadget");
				ret.push(this._sendNotification(notiData, "SMB Missing Devices", "Not all associated devices are detected! Please check if everything is in the bag!"));
				this._bag.setLastGadgetNotificationTs(now);
			} else {
				LOG.info("skipping alert for missing gadget, notification have been already sent", lastNot, now);
			}
		}

		if (ret.length === 0) return;
		return Promise.allSettled(ret)
				then(r => this._saveBagData());
	}

	_storeData(data) {
		let self = this;
		return new Promise((resolve, reject) => {
			if (!data.data || typeof data.data !== "object") {
				return reject(new HandlerException("No valid data field is present!"));
			}

			let record = new BagUpdateDataModel(data.data);
			let validation = record.validate();
			if (validation !== true) {
				return reject(new HandlerException(validation));
			}

			let ts     = Math.floor(record.getTimeStamp() / CONST.DATA_TIME_SLICE) * CONST.DATA_TIME_SLICE;
			let bagKey = self._bag.getKey();

			self._storage
				.appendBagLogs(bagKey, JSON.stringify(record.getData()) + "\n", ts)
				.then(r => self._updateBagData(record))
				.then(r => self._sendNotifications())
				.then(r => resolve(r))
				.catch(e => reject(e));
		});
	}

	_handleError(error) {
		if (error instanceof HandlerException) return this.respondBadRequest(error.message);
		LOG.error("Error during data update", error);
		return this.oops();
	}

	needsMasterAccess() {
		if (this._method.toUpperCase() === "GET") return false;
		return true;
	}
}

module.exports = DataHandler;