const i2c = require('i2c-bus')
const ADS1115 = require('ads1115')
const connection = [1, 0x48, 'i2c-bus']

// Voltage to percentage conversion table
const conv = [
	3.11, 3.14, 3.15, 3.17, 3.19, 3.2, 3.22, 3.24, 3.25, 3.27, 3.29, 3.3, 3.31, 3.32, 3.33, 3.335,
	3.34, 3.35, 3.36, 3.37, 3.38, 3.385, 3.39, 3.4, 3.41, 3.42, 3.43, 3.435, 3.44, 3.45, 3.46,
	3.465, 3.47, 3.48, 3.485, 3.49, 3.5, 3.51, 3.52, 3.525, 3.53, 3.54, 3.545, 3.55, 3.56, 3.565,
	3.57, 3.575, 3.58, 3.585, 3.59, 3.6, 3.605, 3.61, 3.615, 3.62, 3.625, 3.63, 3.64, 3.645, 3.65,
	3.655, 3.66, 3.67, 3.675, 3.68, 3.69, 3.695, 3.7, 3.71, 3.715, 3.72, 3.73, 3.74, 3.745, 3.75,
	3.76, 3.77, 3.775, 3.78, 3.79, 3.8, 3.81, 3.815, 3.82, 3.83, 3.84, 3.85, 3.853, 3.856, 3.86,
	3.865, 3.87, 3.873, 3.876, 3.88, 3.885, 3.89, 3.895, 3.9, 3.93
]

// Offset the conversion table to match average power levels
const voltage_offset = 0.2 //-0.17

module.exports = class Battery {

	constructor (state, terminal) {

		this.state = state	// Save a reference to the global state object
		this.terminal = terminal

		// Init the analog to digital converter board connection
		ADS1115.open(...connection).then(async (ads1115) => {

			/* Set the gain. Valid values:
			 * 2/3 = +/- 6.144V (default)	-> 187.5μV
			 * 1 = +/- 4.096V				-> 125μV
			 * 2 = +/- 2.048V				-> 62.5μV
			 * 4 = +/- 1.024V				-> 31.25μV
			 * 8 = +/- 0.512V				-> 15.625μV
			 * 16 = +/- 0.256V				-> 7.8125μV
			 */
			ads1115.gain = 2/3

			while (true) {

				/* Request a measurement. Valid values for mux parameter:
				 * '0+1' = Differential measurement between A0 & A1
				 * '1+3' = Differential measurement between A1 & A3
				 * '0+3' = Differential measurement between A0 & A3
				 * '2+3' = Differential measurement between A2 & A3
				 * '0+GND' = Single-ended measurement on A0
				 * '1+GND' = Single-ended measurement on A1
				 * '2+GND' = Single-ended measurement on A2
				 * '3+GND' = Single-ended measurement on A3
				 */

				// Request a differential voltage measurement on the two battery poles
				let batt_v = this.intToV(await ads1115.measure('0+1'), 0.0001875)
				let batt_perc = this.vToPercent(batt_v + voltage_offset)

				//console.log('BATT_V - ' + batt_v + ' (' + batt_perc + ' %)')

				// Convert this voltage to percentage and update the state
				this.state.setBatteryCharge(batt_perc)

				// Update terminal indicator
				this.terminal.battery = { v: batt_v, p: batt_perc }

				// Wait 30 seconds before measuring again
				await new Promise(res => setTimeout(res, 10000))
			}
		}).catch((err) => {
			console.log('Battery charge meter disconnected!')
		})
	}

	// Convert a received measurement to voltage
	intToV (num, step) {

		// We receive a signed int interpreted as unsigned, convert it back
		if (num >= 32768) num -= 65536

		// Convert to Volts, round to three decimals
		return Math.round(Math.abs(num) * step * 1000) / 1000
	}

	vToPercent (voltage) {

		// Find the closest value from the conversion table
		for (var p = 0; p <= 100; p++) {
			if (conv[p] > voltage) break
		}

		return Math.min(p, 100)
	}

}
