const path = require('path')
const { exec, execSync } = require('child_process')

// A simple async wrapper for exec
const run = function (command) {
	return new Promise ((resolve, reject) => {
		exec(command, { encoding: 'utf8' }, (err, stdout, stderr) => {
			if (err) {
				reject(err)
				return
			}

			resolve(stdout)
		})
	})
}

const loc_interval = 5000 		// in ms

module.exports = class GPS {

	constructor (state, terminal) {

		this.state = state	// Save a reference to the global state object
		this.terminal = terminal

		this.timer = null

		this.modem = 0

		let modes = this.checkStatus()

		// Initiate the timer for recurring jobs
		this.timer = setInterval(this.getLocation.bind(this), loc_interval)
	}

	// Ask ModemManager for a list of available modems.
	// It is assumed here, that there will be one modem path.
	async getModem () {
		const modem_path_regex = /\s*([^\s]*).*/gmi
		const modems = await run(`mmcli --list-modems`)
		const modem_path = modem_path_regex.exec(modems)[1]
		this.modem = modem_path
	}

	async checkStatus (retries = 0) {

		if (retries > 5) {
			console.log(`GPS module cannot access modem!`)
			return
		}

		try {
			const status = await run(`mmcli --modem=${this.modem} --location-status`)
			const status_regex = /enabled: (.*)\n/gmi
			const gps_modes = status_regex.exec(status)[1].split(', ')

			console.log(`gps_modes: ${gps_modes.join(', ')}`)

			if (gps_modes.length < 3) {
				// Not all gps modes are active, let's enable them
				await run(`mmcli --modem=${this.modem} --location-enable-gps-raw`)
				await run(`mmcli --modem=${this.modem} --location-enable-gps-nmea`)
				//await run(`mmcli --modem=${this.modem} --location-enable-agps`)
				await run(`mmcli --modem=${this.modem} --location-set-gps-refresh-rate=5`)
			}

		} catch (err) {
			if (err.message.includes(`couldn't find modem`)) {
				// Wait a second
				await new Promise(res => setTimeout(res, 1000))
				// Reacquire modem
				await this.getModem()
				// Retry
				await this.checkStatus(retries + 1)
			} else {
				// Other unexpected error
				console.log(`GPS - unexpected error: ${err.message}`)
			}
		}
	}

	async getLocation () {

		// First ensure that the modem is accessible and the gps modes are all active
		await this.checkStatus()

		try {
			// Now get the latest info
			const location_data = await run(`mmcli --modem=${this.modem} --location-get`)

			const longitude_regex = /longitude:\s*([^\s]*)\s*$/gm
			const latitude_regex = /latitude:\s*([^\s]*)\s*$/gm
			const dop_regex = /\$GNGSA.*\,([^\,]*)\,([^\,]*)\,([^\,]*)\,[^\,]*$/gm

			let res = null
			let lon = null
			let lat = null
			let hdop = null
			let vdop = null
			let dop = null
			if ((res = longitude_regex.exec(location_data)) !== null) {
				lon = res[1]
			}
			if ((res = latitude_regex.exec(location_data)) !== null) {
				lat = res[1]
			}
			if ((res = dop_regex.exec(location_data)) !== null) {
				hdop = res[1]
				vdop = res[2]
				if (hdop != '' && vdop != '') {
					dop = Math.max(parseFloat(hdop), parseFloat(vdop))
				}
			}

			if (lon !== null && lat !== null && dop !== null) {
				// We have a gps fix!
				console.log(`gps fix: ${lat} ${lon} dop: ${dop} (${dop <= 2 ? 'Excellent' : dop <= 5 ? 'Good' : dop <= 10 ? 'Moderate' : dop <= 20 ? 'Fair' : 'Poor DISCARDED' })`)

				// Send the location to the global state object
				if (dop <= 20) this.state.setLocation(lat, lon, 0)
			} else {
				console.log('no gps fix')
			}

		} catch (err) {
			if (err.message.includes(`couldn't find modem`)) {
				console.log(`GPS - modem lost, reacquiring`)
			} else {
				console.log(`GPS - unexpected error: ${err.message}`)
			}
		}
	}
}
