var terminal_enabled = false
process.argv.forEach((val, index, array) => {
	if (val == '-debug') terminal_enabled = true
})

const tty = require('tty')
const min_width = 80
const min_height = 30

const stdout = process.stdout
const stdin = process.stdin

const beacon_colors = [ '', '\x1b[30m', '\x1b[37m', '\x1b[36m', '\x1b[34m', '\x1b[32m', '\x1b[31m' ]
								// In order: unspecified, black, white, cyan, blue, green, red

const beacon_distances = [ 'near', 'far ', 'lost' ]

class Terminal {

	constructor () {
		// To activate the console, run the program with the '-debug' parameter
		if (stdout.isTTY && terminal_enabled) {

			// Device info
			this._processor_id = '-'
			this._hci_peripheral = '-'
			this._hci_central = '-'
			this._hci_scanning = false
			this._ble_conn = false
			this._tunnel_conn = false

			// Cellular
			this._cell_uart = '-'
			this._cell_state = '-'
			this._hotspot = {
				on: false,
				ssid: '-',
				pass: '-',
			}

			// Battery
			this._batt_conn = 'Not connected'
			this._batt_v = 0
			this._batt_perc = 0

			// Beacons
			this._beacons = {}

			this.pipeline = []
			this.rendering = false

			// Overload console.log ----------------------------------------------------------------

			this.buffer = []
			this.lines = []
			this.offset = 0
			this.offset_max = 0
			this.c_columns = stdout.columns - 53
			this.c_rows = stdout.rows - 3
			this.wrap_regex = new RegExp(`(?![^\\n]{1,${this.c_columns}}$)([^\\n]{1,${this.c_columns}})`, 'gm')
			this.wrap = (s) => s.replace(this.wrap_regex, '$1\n')

			// Save a reference to the original
			this.console_log = console.log
			this.console_err = console.err

			// Replace the function
			console.log = this.log.bind(this)
			console.err = this.log.bind(this)

			// Output handler ----------------------------------------------------------------------

			stdout.on('resize', () => {
				this.rewriteLineBuffer()
				this.render('layout')
			})
			this.render('layout')

			// Clock timer
			this.clock = setInterval(() => { this.render('clock') }, 1000)

			// Input handler -----------------------------------------------------------------------

			// Readline lets us tap into the process events
			const readline = require('readline')

			// Allows us to listen for events from stdin
			readline.emitKeypressEvents(process.stdin)

			// Raw mode gets rid of standard keypress events and other
			// functionality Node.js adds by default
			stdin.setRawMode(true)

			// Start the keypress listener for the process
			stdin.on('keypress', (str, key) => {

				// "Raw" mode so we must do our own kill switch
				if (key.sequence === '\u0003') {
					stdout.cursorTo(0, 0)
					this.setColor('reset')
					stdout.clearScreenDown()
					this.showCursor()
					process.exit()
				}

				if (key.name == 'home') {
					this.offset = 0
					this.render('console')
				}

				if (key.name == 'end') {
					this.offset = this.offset_max
					this.render('console')
				}

				if (key.name == 'pageup') {
					this.offset = Math.max(0, this.offset - this.c_rows)
					this.render('console')
				}

				if (key.name == 'pagedown') {
					this.offset = Math.min(this.offset_max, this.offset + this.c_rows)
					this.render('console')
				}

				if (key.name == 'up') {
					this.offset = Math.max(0, this.offset - 1)
					this.render('console')
				}

				if (key.name == 'down') {
					this.offset = Math.min(this.offset_max, this.offset + 1)
					this.render('console')
				}
			})
		}
	}

	log (...params) {
		// Concat output and add data to buffer
		params.forEach((el, i) => {
			if (typeof el !== 'string') {
				params[i] = JSON.stringify(el, null, '    ')
			}
		})
		let entry = params.join(' ')
		this.buffer.push(entry)

		// Go easy on the memory
		if (this.buffer.length > 1000) {
			this.buffer.shift()
		}

		// Line buffer
		let c_rows = stdout.rows - 3
		this.lines.push(...(this.wrap(entry).split('\n')))

		// Offset
		if (this.offset == this.offset_max) {
			// Stick to the end
			this.offset_max = Math.max(0, this.lines.length - this.c_rows)
			this.offset = this.offset_max
		} else {
			// Keep it where it was
			this.offset_max = Math.max(0, this.lines.length - this.c_rows)
			this.offset = Math.min(this.offset_max, this.offset)
		}

		// Redraw output
		this.render('console')
	}

	rewriteLineBuffer () {
		this.c_columns = stdout.columns - 53
		this.c_rows = stdout.rows - 3
		this.wrap_regex = new RegExp(`(?![^\\n]{1,${this.c_columns}}$)([^\\n]{1,${this.c_columns}})`, 'gm')

		this.lines = []
		for (let i = 0; i < this.buffer.length; i++) {
			let entry = this.buffer[i]
			this.lines.push(...(this.wrap(entry).split('\n')))
		}
		this.offset_max = Math.max(0, this.lines.length - this.c_rows)
		this.offset = Math.min(this.offset_max, this.offset)
	}

	// Setters -------------------------------------------------------------------------------------

	set processor_id (value) {
		if (terminal_enabled) {
			this._processor_id = value
			this.render('deviceInfo')
		}
	}

	set hci_peripheral (value) {
		if (terminal_enabled) {
			this._hci_peripheral = value
			this.render('deviceInfo')
		}
	}

	set hci_central (value) {
		if (terminal_enabled) {
			this._hci_central = value
			this.render('deviceInfo')
		}
	}

	set hci_scanning (value) {
		if (terminal_enabled) {
			this._hci_scanning = value
			this.render('deviceInfo')
		}
	}

	set ble_conn (value) {
		if (terminal_enabled) {
			this._ble_conn = value
			this.render('deviceInfo')
		}
	}

	set tunnel_conn (value) {
		if (terminal_enabled) {
			this._tunnel_conn = value
			this.render('deviceInfo')
		}
	}

	set cell_uart (value) {
		if (terminal_enabled) {
			this._cell_uart = value
			this.render('cellular')
		}
	}

	set cell_state (value) {
		if (terminal_enabled) {
			this._cell_state = value
			this.render('cellular')
		}
	}

	set hotspot (value) {
		if (terminal_enabled) {
			this._hotspot = {
				on: value.on,
				ssid: value.ssid,
				pass: value.pass,
			}
			this.render('cellular')
		}
	}

	set battery (value) {
		if (terminal_enabled) {
			this._batt_conn = 'OK'
			this._batt_v = value.v
			this._batt_perc = value.p
			this.render('battery')
		}
	}

	set beacons (value) {
		if (terminal_enabled) {
			this._beacons = value
			this.render('beacons')
		}
	}

	// Renderers -----------------------------------------------------------------------------------

	async render (...new_jobs) {

		// Enqueue any new render jobs
		if (new_jobs.length > 0) {
			this.pipeline.push(...new_jobs)
		}

		// Make sure this only happens in this thread
		if (this.rendering) return
		this.rendering = true

		// Handle all pending jobs
		while (this.pipeline.length > 0) {
			let job = this.pipeline.shift()
			if (job != '') {
				this['print' + job.charAt(0).toUpperCase() + job.slice(1)]()
			}
		}

		// Release
		this.rendering = false
	}

	printLayout () {

		this.hideCursor()

		if (stdout.rows < min_height || stdout.columns < min_width) {
			this.setColor('reset')
			stdout.cursorTo(0, 0)
			stdout.clearScreenDown()
			stdout.write('Terminal window is too small.')
			return
		}

		// Header
		this.setColor('header')
		stdout.cursorTo(0, 0)
		stdout.clearLine()
		stdout.write(' \u25a0 Smart Modding Bag')
		this.printClock()

		// Frames
		this.setColor('info-frame')
		stdout.cursorTo(0, 1)
		stdout.clearScreenDown()
		stdout.write('\u2554')
		this.horizLine(49, '\u2550')
		stdout.write('\u2557')
		this.setColor('console-frame')
		stdout.write('\u2554')
		this.horizLine(stdout.columns - 53, '\u2550')
		stdout.write('\u2557')

		let separators = [ 7, 17, 23 ]

		for (let y = 2; y < stdout.rows - 1; y++) {
			this.setColor('info-frame')
			stdout.cursorTo(0, y)
			if (separators.includes(y)) {
				stdout.write('\u255f')
				this.horizLine(49, '\u2500')
				stdout.write('\u2562')
			} else {
				stdout.write('\u2551')
				stdout.cursorTo(50, y)
				stdout.write('\u2551')
			}
			this.setColor('console-frame')
			stdout.write('\u2551')
			this.setColor('console')
			stdout.clearLine(1)

			//stdout.cursorTo(stdout.columns - 1, y)
			//stdout.write('\u2551')
		}

		this.setColor('info-frame')
		stdout.cursorTo(0, stdout.rows - 1)
		stdout.write('\u255a')
		this.horizLine(49, '\u2550')
		stdout.write('\u255d')
		this.setColor('console-frame')
		stdout.write('\u255a')
		this.horizLine(stdout.columns - 53, '\u2550')
		stdout.write('\u255d')

		// Frame titles
		stdout.cursorTo(53, 1)
		stdout.write(' Console ')

		this.setColor('info-frame')
		stdout.cursorTo(2, 1)
		stdout.write(' Device info ')
		stdout.cursorTo(2, 7)
		stdout.write(' Cellular ')
		stdout.cursorTo(2, 17)
		stdout.write(' Battery ')
		stdout.cursorTo(2, 23)
		stdout.write(' Beacons ')

		// Indicators
		this.printDeviceInfo()
		this.printCellular()
		this.printBattery()
		this.printBeacons()
		this.printConsole()
	}

	printClock () {
		this.setColor('header')
		stdout.cursorTo(stdout.columns - 21, 0)
		stdout.write(new Date().toString().split(' ').slice(1, 5).join(' '))
	}

	printDeviceInfo () {
		this.setColor('info')
		stdout.cursorTo(2, 3)
		stdout.write('SMB Serial:      ' + this._processor_id + '  ' + (this._hci_scanning ? '\x1b[32mSCAN\x1b[37m  ' : '      ') + (this._ble_conn ? '\x1b[32mBLE\x1b[37m  ' : '     ') + (this._tunnel_conn ? '\x1b[32mTUNNEL\x1b[37m' : '      '))
		stdout.cursorTo(2, 4)
		stdout.write('HCI Peripheral:  ' + this._hci_peripheral.replace('UNDETECTED', '\x1b[48;5;88mUNDETECTED\x1b[48;5;235m'))
		stdout.cursorTo(2, 5)
		stdout.write('HCI Central:     ' + this._hci_central.replace('UNDETECTED', '\x1b[48;5;88mUNDETECTED\x1b[48;5;235m'))
	}

	printCellular () {
		this.setColor('info')
		stdout.cursorTo(2, 9)
		stdout.write('UART:            ' + this._cell_uart.padEnd(20))
		stdout.cursorTo(2, 10)
		stdout.write('State:           ' + this._cell_state.padEnd(20))
		stdout.cursorTo(2, 11)
		stdout.write('PIN:             -')
		stdout.cursorTo(2, 12)
		stdout.write('PPP:             -')
		stdout.cursorTo(2, 13)
		stdout.write('GPS Location:    -')
		stdout.cursorTo(2, 14)
		stdout.write('GSM Location:    -')
		stdout.cursorTo(2, 15)
		stdout.write('Hotspot:         ' + (this._hotspot.on ? 'ON  ' : 'OFF  ' + this._hotspot.ssid + '  ' + this._hotspot.pass).padEnd(30))
	}

	printBattery () {
		this.setColor('info')
		stdout.cursorTo(2, 19)
		stdout.write('ADC:             ' + this._batt_conn.padEnd(20))
		stdout.cursorTo(2, 20)
		stdout.write('Battery voltage: ' + (this._batt_v + 'V').padEnd(20))
		stdout.cursorTo(2, 21)
		stdout.write('Battery charge:  ' + (this._batt_perc + '%').padEnd(20))
	}

	printBeacons () {
		this.setColor('info')
		stdout.cursorTo(2, 25)
		stdout.write('Paired:                                      ')
		stdout.cursorTo(2, 26)

		let offset = 0
		Object.keys(this._beacons).forEach((address) => {
			let b = this._beacons[address]
			if (b.p == 1) {
				stdout.cursorTo(2, 26 + offset)
				stdout.write('  ' + beacon_colors[b.c] + '\u25b2\x1b[37m  ' + address + '  ' + new Date(b.l).toString().split(' ')[4] + '  ' + beacon_distances[b.d])
				offset++
			}
		})
		if (offset == 0) {
			// Nothing was paired
			stdout.write('  -                                          ')
			offset++
		}

		stdout.cursorTo(2, 26 + offset)
		stdout.write('                                             ')
		offset++
		stdout.cursorTo(2, 26 + offset)
		stdout.write('Seen:                                        ')
		offset++
		stdout.cursorTo(2, 26 + offset)

		let offset2 = offset
		Object.keys(this._beacons).forEach((address) => {
			let b = this._beacons[address]
			if (b.p == 0) {
				stdout.cursorTo(2, 26 + offset)
				stdout.write('  ' + beacon_colors[b.c] + '\u25b2\x1b[37m  ' + address + '  ' + new Date(b.l).toString().split(' ')[4] + '  ' + beacon_distances[b.d])
				offset++
			}
		})
		if (offset == offset2) {
			// Nothing was seen
			stdout.write('  -                                          ')
			offset++
		}

		// Keep clearing the screen until we hit bottom
		for (offset; offset < stdout.rows - 27; offset++) {
			stdout.cursorTo(2, 26 + offset)
			stdout.write('                                             ')
		}
	}

	printConsole () {

		// Print the lines
		this.setColor('console')
		let offset = Math.max(this.lines.length - this.c_rows, 0)
		let line = null
		for (let row = 0; row < this.c_rows && this.offset + row < this.lines.length; row++) {
			// Print a line to the console
			stdout.cursorTo(52, 2 + row)
			stdout.write(this.lines[this.offset + row].padEnd(this.c_columns, ' '))
		}

		// Redraw the scrollbar
		let bar_height = (this.lines.length == 0) ? (this.c_rows - 2) : Math.min((this.c_rows - 2), Math.max(1, Math.round((this.c_rows - 2) * this.c_rows / this.lines.length)))
		let bar_start = 3 + ((this.offset_max == 0) ? 0 : this.offset * ((this.c_rows - 2) - bar_height) / this.offset_max)
		let bar_end = bar_start + bar_height

		this.setColor('scrollbar')
		stdout.cursorTo(stdout.columns - 1, 2)
		stdout.write('\u25b2')
		for (let y = 3; y < stdout.rows - 2; y++) {
			stdout.cursorTo(stdout.columns - 1, y)
			if (y >= bar_start && y < bar_end) {
				//stdout.write('\u2593')
				stdout.write('\u2588')
			} else {
				//stdout.write('\u2591')
				stdout.write(' ')
			}
		}
		stdout.cursorTo(stdout.columns - 1, stdout.rows - 2)
		stdout.write('\u25bc')
	}

	// Terminal control ----------------------------------------------------------------------------

	horizLine (len, char) {
		for (let x = 0; x < len; x++) {
			stdout.write(char)
		}
	}

	hideCursor () {
		stdout.write('\x1b[0m\x1b[?12l\x1b[?25l')
	}

	showCursor () {
		stdout.write('\x1b[0m\x1b[?12h\x1b[?25h')
	}

	setColor (color) {
		switch (color) {
			case 'reset':
				stdout.write('\x1b[0m')
				break
			case 'header':
				stdout.write('\x1b[48;5;166m\x1b[30m')
				break
			case 'info-frame':
				stdout.write('\x1b[48;5;235m\x1b[38;5;166m')
				break
			case 'console-frame':
				stdout.write('\x1b[40m\x1b[38;5;243m')
				break
			case 'console':
				stdout.write('\x1b[40m\x1b[37m')
				break
			case 'scrollbar':
				stdout.write('\x1b[48;5;235m\x1b[37m')
				break
			case 'info':
				stdout.write('\x1b[48;5;235m\x1b[37m')
				break
		}
	}
}

module.exports = new Terminal()
