'use strict';
const CONST = require("../constants.js");
const LOG   = require("../core/Logger.js");

const ServerMessageBuilder = require("./messages/ServerMessageBuilder.js");

const MESSAGE_TYPE_OPEN_CHANNEL="open";
const MESSAGE_TYPE_PING="ping";
const MESSAGE_TYPE_SEND_MESSAGE="send";

class Client {
	constructor(id, socket, config, clientStore) {
		//id assigned by the client store
		this._id     = id;
		this._socket = socket;
		this._config = config;
		this._clientStore = clientStore;
		this._channel = null;
		this._closing = false;
		this._logPrefix = "|" + id + "|";

		this._closeWaitTime = this._config.get("tunnel.close_wait_time");
		this._channelOpenTimeout = this._config.get("tunnel.channel_open_timeout");
		this._pingInterval = this._config.get("tunnel.ping_interval", 10000);
		this._clientTimeout = this._config.get("tunnel.client_timeout", 30000);

		//this is the ID provided by the client in channel open requests
		this._callId = null;
		this._partnerCallId = null;
		this._lastMessage = 0;
		this._healthChecker = null;

		this._startMessageHandling();
	}

	_updateLastMessage() {
		this._lastMessage = Date.now();
	}

	_startMessageHandling() {
		this._socket.on("message", message => this._handle(message));
		this._socket.on("close", event => this._handleClose(event));
		this._socket.on("error", event => this._handleError(event));
		this._socket.on("ping", event => this._handlePingPong("PING"));
		this._socket.on("pong", event => this._handlePingPong("PONG"));

		this._healthChecker = setInterval(() => {
			if (Date.now() - this._lastMessage > this._clientTimeout) {
				LOG.error(this._logPrefix, "No packet was received from client within the timeout period, closing connection!");
				this.close(1008, "No packet was received from client within the timeout period, closing connection!");
			} else {
				this._socket.ping(() => {});
			}
		}, this._pingInterval);

		//Self close if no channel gets opened within the grace period
		setTimeout(() => {
			if (!this._channel) this.close(1008, "No channel open request received within the grace period.");
		}, this._channelOpenTimeout);
	}

	_handlePingPong(event) {
		if (this._closing) {
			LOG.debug(this._logPrefix, "Socket is closing, ignoring messages");
			return;
		}

		LOG.debug(this._logPrefix, event, "event received");
		this._updateLastMessage();
	}

	_handle(message) {
		if (this._closing) {
			LOG.debug(this._logPrefix, "Socket is closing, ignoring messages");
			return;
		}

		LOG.debug(this._logPrefix, "Message received", message);
		this._updateLastMessage();
		let msg = {};
		try {
			msg = JSON.parse(message);
		} catch (e) {
			//We'll handle invalid messages later
		}

		switch (msg.type) {
			case MESSAGE_TYPE_OPEN_CHANNEL:
				this._messageOpenChannel(msg);
			break;
			case MESSAGE_TYPE_PING:
				this._messagePing(msg);
			break;
			case MESSAGE_TYPE_SEND_MESSAGE:
				this._messageSendMessage(msg);
			break;
			default:
				this._invalidMessage("Invalid message received, terminating connection. '" + message + "'")
				this.close(1007, "Invalid message!");
		}
	}

	_handleError(event) {
		LOG.error(this._logPrefix, "Error occured, closing connection", event);
		this.close(1011, "Communication error, see server logs for details.");
	}

	_handleClose(event) {
		LOG.info(this._logPrefix, "Socket has been closed clean");
		//if the close was clean we can clear the terminate timeout
		if (this._cleanupTimeout) clearTimeout(this._cleanupTimeout);
		this._cleanup();
	}

	_invalidMessage(logReason) {
		LOG.error(this._logPrefix, logReason);
		this.close(1007, "Invalid message received");
	}

	_messageOpenChannel(msg) {
		let {me, rec, seq} = msg;

		if (typeof me !== "string" || me.length > CONST.MAX_ID_LENGTH) {
			this._invalidMessage("Invalid sender identifier for open request '" + me + "'");
		}

		if (typeof rec !== "string" || rec.length > CONST.MAX_ID_LENGTH) {
			this._invalidMessage("Invalid recipient identifier for open request '" + rec + "'");
		}

		if (typeof seq !== "number") {
			this._invalidMessage("Invalid squence identifier '" + seq + "'");
		}

		if (this._channel) {
			this.send(ServerMessageBuilder
				.message(ServerMessageBuilder.MESSAGE_TYPE_ACK, this._channel)
				.setSequence(seq)
				.setError(1, "Already connected to a channel")
				.build()
			);

			return
		}

		this._callId = me;
		this._partnerCallId = rec;

		this._channel = this._clientStore.provideChannel(this._callId, this._partnerCallId, this);

		this.send(ServerMessageBuilder
			.message(ServerMessageBuilder.MESSAGE_TYPE_ACK, this._channel)
			.setSequence(seq)
			.build()
		);

	}

	_messagePing(msg) {
		let {seq} = msg;

		if (typeof seq !== "number") {
			this._invalidMessage("Invalid squence identifier '" + seq + "'");
		}

		this.send(ServerMessageBuilder
			.message(ServerMessageBuilder.MESSAGE_TYPE_ACK, this._channel)
			.setSequence(seq)
			.build()
		);
	}

	_messageSendMessage(msg) {
		let {seq, d} = msg;

		if (typeof seq !== "number") {
			this._invalidMessage("Invalid squence identifier '" + seq + "'");
		}

		if (!this._channel) {
			this.send(ServerMessageBuilder
				.message(ServerMessageBuilder.MESSAGE_TYPE_ACK, this._channel)
				.setSequence(seq)
				.setError(1, "Already connected to a channel")
				.build()
			);

			return
		}

		if (!this._channel.hasClient(this._partnerCallId)) {
			this.send(ServerMessageBuilder
				.message(ServerMessageBuilder.MESSAGE_TYPE_ACK, this._channel)
				.setSequence(seq)
				.setError(1, "Already connected to a channel")
				.build()
			);

			return
		}

		this._channel.sentTo(this._partnerCallId, ServerMessageBuilder
			.message(ServerMessageBuilder.MESSAGE_TYPE_DATA, this._channel)
			.setData(d)
			.build()
		);

		this.send(ServerMessageBuilder
			.message(ServerMessageBuilder.MESSAGE_TYPE_ACK, this._channel)
			.setSequence(seq)
			.build()
		);
	}

	_cleanup() {
		this._clientStore.removeClient(this._id);
		if (this._channel) this._channel.removeClient(this);
		if (this._healthChecker) clearInterval(this._healthChecker);
	}

	send(message) {
		LOG.debug(this._logPrefix, "Out message:", message);
		this._socket.send(message);
	}

	getCallId() {
		return this._callId;
	}

	getPartnerCallId() {
		return this._partnerCallId;
	}

	getChannel() {
		return this._channel;
	}

	close(code, reason) {
		this._closing = true;
		this._socket.close(code, reason);
		LOG.info(this._logPrefix, "Client connection closed: ", reason);
		this._cleanupTimeout = setTimeout(() => {
			if ([this._socket.OPEN, this._socket.CLOSING].includes(this._socket.readyState)) {
				this._socket.terminate();
			}
		}, this._closeWaitTime);

		this._cleanup();
	}
}

module.exports = Client;