'use strict';
const CONST = require("../constants.js");
const LOG   = require("../core/Logger.js");

class APIRequestHandlerBase {
	constructor(request, vars, context) {
		this._request   = request;
		this._vars      = vars || {};
		this._body      = this._request.body || "";
		this._method    = this._request.method;
		this._headers   = this._request.headers;
		this._path      = request.url;
		this._requestId = request.id;
		this._dal       = context.dal;
		this._mail      = context.mail;
		this._cookies      = this._parseCookies(request);
		this._writeCookies = {};
		this._crypto       = context.crypto;

		this._logPrefix = "|#" + request.id + "|";
		this._startTime = Date.now();
		this._externalHost = context.externalHost;
	}

	static getRoute() {
		throw new Error("Method getRoute must be implemented!");
	}

	respondOk(body) {
		return this.buildOkResponse(200, body);
	}

	respondCreated(body) {
		return this.buildOkResponse(201, body);
	}

	respondAccepted(body) {
		return this.buildOkResponse(202, body);
	}

	redirect(url, body) {
		return this.buildOkResponse(301, body || "", {"Location" : url});
	}

	redirectTemp(url, body) {
		return this.buildOkResponse(302, body || "", {"Location" : url});
	}

	respondBadRequest(err) {
		return this.buildErrorResponse(400, err);
	}

	respondUnauthorized(err) {
		return this.buildErrorResponse(401, err);
	}

	respondForbidden(err) {
		return this.buildErrorResponse(403, err);
	}

	respondNotFound(err) {
		return this.buildErrorResponse(404, err);
	}

	respondMethodNotAllowed(method) {
		return this.buildErrorResponse(405, "HTTP Method '" + method + "' not allowed!", "server.method_not_allowed");
	}

	respondConflict(err) {
		return this.buildErrorResponse(409, err);
	}

	respondInternalServerError() {
		return this.buildErrorResponse(500, "Internal server error");
	}

	oops() {
		return this.respondInternalServerError();
	}

	buildErrorResponse(code, err, headers) {
		let response = this.buildResponse(code, {
			ok      : false,
			error   : {
				reason : err
			},
			ver     : CONST.PROTOCOL_VERSION,
			ver_min : CONST.PROTOCOL_COMPATIBLITY_VERSION,
		}, headers);
		response.errorResponse = true;
		return response;
	}

	buildOkResponse(code, body, headers) {
		let core     = {
			...body,
			ok      : true,
			ver     : CONST.PROTOCOL_VERSION,
			ver_min : CONST.PROTOCOL_COMPATIBLITY_VERSION
		};

		return this.buildResponse(code, core, headers);
	}

	buildResponse(code, body, headers) {
		let ret = {
			code : code,
			body : typeof body === "object" ? JSON.stringify(body) : body
		}

		let cookieHeaders = this._buildCookieHeader();
		if (cookieHeaders.length > 0) {
			headers = headers || {};
			headers["Set-Cookie"] = cookieHeaders;
		}

		if (headers) ret.headers = headers;

		return ret;
	}

	async _isAuthenticated() {
		return false;
	}

	async _hasAccess() {
		return false;
	}

	/***
	 * Instead of overwriting the constructor, member variables should be initialized in this method
	 */
	async _init() {
	}

	/***
	 * Object initializations and DB connections should be here, this is called after the authentication data check
	 */
	async _setup() {
	}

	/***
	 * This is the main process method, which should return the result on finish
	 */
	async _process() {
		let method = this._method;
		let ret;
		switch (method.toUpperCase()) {
			case "GET":
				ret = this._processGET();
			break;
			case "PUT":
				ret = this._processPUT();
			break;
			case "POST":
				ret = this._processPOST();
			break;
			case "DELETE":
				ret = this._processDELETE();
			break;
			default:
				return this.respondMethodNotAllowed(method);
			break;
		}

		return ret.catch(e => this._processError(e));
	}

	async _processGET() {
		return this.respondMethodNotAllowed("GET");
	}

	async _processPUT() {
		return this.respondMethodNotAllowed("PUT");
	}

	async _processPOST() {
		return this.respondMethodNotAllowed("POST");
	}

	async _processDELETE() {
		return this.respondMethodNotAllowed("DELETE");
	}

	_processError(e) {
		if (e && e.errorResponse) {
			delete e.errorResponse;
			return e;
		}
		LOG.error(this._logPrefix, "Unhandled error catched in request handler: ", e);
		return this.respondInternalServerError("Internal server error!");
	}

	_setCookie(name, value, ttl, path) {
		if (ttl) {
			value += "; Max-Age=" + ttl;
		}

		if (path) {
			value += "; Path=" + path;
		}

		this._writeCookies[name] = value;
	}

	_parseCookies(request) {
		let ret = {};
		let cookieHeader = request.headers ? (request.headers.cookie || "") : "";
		let cookies = cookieHeader.split(";");

		for (let i = 0; i < cookies.length; i++) {
			let parts = cookies[i].split("=");
			parts[0] = parts[0].trim();
			if (parts.length !== 2 || !parts[0]) continue;
			ret[parts[0]] = parts[1];
		}

		return ret;
	}

	_buildCookieHeader() {
		let cookies = [];
		for (let i in this._writeCookies) {
			cookies.push(i + "=" + this._writeCookies[i]);
		}
		return cookies;
	}

	async handle() {
		try {
			await this._init();
		} catch (e) {
			return e;
		}

		if (!await this._isAuthenticated()) {
			return this.respondUnauthorized("You are not authenticated!");
		}

		try {
			await this._setup();
		} catch (e) {
			return e;
		}

		if (!await this._hasAccess()) {
			return this.respondForbidden("You have no access to this resource!");
		}

		return this._process();
	}
}

module.exports = APIRequestHandlerBase;