'use strict';
const CONST  = require("../../constants.js");
const LOG    = require("../../core/Logger.js");

const APIRequestHandlerBase = require("../APIRequestHandlerBase.js");
const HandlerException = require("./HandlerException.js");

class UserBasedHandlerBase extends APIRequestHandlerBase {
	async _init() {
		if (this._body && this._headers["content-type"] === "application/json") {
			try {
				this._body = JSON.parse(this._body);
			} catch (e) {
				LOG.info("Invalid JSON received", this._body, e);
				return Promise.reject(this.respondBadRequest("Invalid JSON payload"));
			}
		}

		try {
			this._auth = this._getAuth();
		} catch (e) {
			LOG.info("Invalid Authentication received received", this._cookies, e);
			return Promise.reject(this.respondBadRequest("Invalid AUTH data"));
		}

		if (this._body && typeof(this._body) !== "object") {
			LOG.info("Payload is not JSON, or header('" + this._headers["content-type"] + "') is missing", this._body);
			return Promise.reject(this.respondBadRequest("Invalid request payload, should be JSON and content-type should be provided as well."));
		}
	}

	_getAuth() {
		let ret = null
		let auth = this._cookies["smbuiauth"];

		if (!auth) return null;
		ret = JSON.parse(this._crypto.decrypt(auth));

		//we check if we have a user id and it is the same id as the url we are trying to access
		//we check if we have a ttl, and it is not expired yet
		if (!ret.id || ret.id !== this._vars.userId || !ret.ttl || ret.ttl < this._startTime) {
			return null;
		}

		ret.ttl = this._startTime + CONST.AUTH_TTL_SEC * 1000;
		return ret;
	}

	async _setup() {
		return this._dal.getUserData(this._vars.userId)
			.then(res => {
				this._user = res;
			}).catch(e => Promise.reject(this.respondNotFound("Unknown user")));
	}

	_refreshAuthCookie() {
		this._setCookie("smbuiauth", this._crypto.encrypt(JSON.stringify(this._auth)), CONST.AUTH_TTL_SEC, "/");
		this._setCookie("smbuiuser", this._auth.id, CONST.AUTH_TTL_SEC, "/");
	}

	async _isAuthenticated() {
		this._refreshAuthCookie();
		return this._auth ? true : false;
	}

	async _hasAccess() {
		return this._user.isActivated();
	}

	_saveUserData() {
		return this._dal.updateUserData(this._user.getId(), this._user.getData());
	}
}

module.exports = UserBasedHandlerBase;