'use strict';

const CONST  = require("../../constants.js");
const LOG    = require("../../core/Logger.js");

const APIRequestHandlerBase = require("../APIRequestHandlerBase.js");
const HandlerException = require("./HandlerException.js");
const UserModel = require("../dal/UserModel.js");

class UserRegistrationHandler extends APIRequestHandlerBase {
	static getRoute() {
		return "/user/register";
	}

	async _init() {
		if (this._body && this._headers["content-type"] === "application/json") {
			try {
				this._body = JSON.parse(this._body);
			} catch (e) {
				LOG.info("Invalid JSON received", this._body, e);
				return Promise.reject(this.respondBadRequest("Invalid JSON payload"));
			}
		}

		if (this._body && typeof(this._body) !== "object") {
			LOG.info("Payload is not JSON, or header('" + this._headers["content-type"] + "') is missing", this._body);
			return Promise.reject(this.respondBadRequest("Invalid request payload, should be JSON and content-type should be provided as well."));
		}
	}

	async _processGET() {
		return this.respondMethodNotAllowed("GET");
	}

	async _processPUT() {
		return this.respondMethodNotAllowed("PUT");
	}

	async _processDELETE() {
		return this.respondMethodNotAllowed("DELETE");
	}

	async _processPOST() {
		if (!this._body) {
			return this.respondBadRequest("Missing data fields!");
		}

		if (!this._body.email || !this._body.clientToken || !this._body.password || !this._body.bagId || !this._body.name) {
			return this.respondBadRequest("Missing data fields!");
		}

		if (!this._body.email.match(/^[^@\s]+@[^@.\s]+(.[^@.\s]+)+$/)) {
			return this.respondBadRequest("Invalid email format!");
		}

		this._user = new UserModel(this._body.email);
		this._user.parseClientData(this._body);
		let token = Math.round(Math.random() * 60466176).toString(36).toUpperCase();
		this._user.setActivationToken(token);

		return this._dal.getUserData(this._body.email)
			.then(r => this.respondBadRequest("User already exists"))
			.catch(r => {
				return this._saveUserData()
					.then(r => this._notifyUser(token))
					.then(r => this.respondOk({
						data: this._user.getClientData()
					}))
					.catch(e => this._handleError(e));
			})
	}

	_notifyUser(token) {
		return this._mail.sendMail({
			to : this._user.getId(),
			subject: "Activation link for SMB account",
			text: "Please visit the following link: " + this._externalHost + "/user/" + this._user.getId() + "/activate/"  + token
		});
	}

	_handleError(error) {
		if (error instanceof HandlerException) return this.respondBadRequest(error.message);
		LOG.error("Error during user update", error);
		return this.oops();
	}

	async _isAuthenticated() {
		return true;
	}

	async _hasAccess() {
		return true;
	}

	_saveUserData() {
		return this._dal.updateUserData(this._user.getId(), this._user.getData());
	}
}

module.exports = UserRegistrationHandler;