'use strict';

class ModelBase {
	constructor(data, dataPipe, dataFactory, options) {
		this._dataPipe    = dataPipe;
		this._rawData     = data;
		this._dataFactory = dataFactory;
		this._options     = options || {};

		this._data     = data ? this._parseData(data) : this._initialData();
		this._create   = !this.getId();
		this._afterParseData();
	}

	updateRawData(data) {
		this._data = this._parseData(data);
		this._afterParseData();
	}

	setDataFactory(factory) {
		this._dataFactory = factory;
	}

	save(options) {
		let saveData = this._prepareData();

		if (this._create) {
			return this._dataPipe.createData(saveData, options).then(result => {
				let [id, data, options] = result;

				this._data = this._parseData(data);
				this._afterParseData();

				if (id) this._setId(id);
				this._create = false;
				return this;
			});
		} else {
			return this._dataPipe.saveData(this.getId(), saveData, options).then(result => {
				let [id, data, options] = result;

				this._data = this._parseData(data);
				this._afterParseData();

				return this;
			});
		}
	}

	_setId(id) {
		this._data.id = id;
	}

	getId() {
		return this._data ? this._data.id : undefined;
	}

	getData() {
		return {...this._data};
	}

	/***
	 * This method is invoked after the _data field has been set either from parsing data or getting it from constructor
	 */
	_afterParseData() {
	}


	/***
	 * Data conversion method, it converts INTERNAL data format to EXTERNAL
	 * @return {mixed} - External format of our data
	 */
	_prepareData() {
		return this._data;
	}

	/***
	 * Data conversion method, it converts EXTERNAL data format to INTERNAL
	 * @param {mixed} data - The external format of the data
	 * @return {mixed} - Internal format of our data
	 */
	_parseData(data) {
		return data;
	}

	/***
	 * This method is invoked when no data is provided, and we need to initialize an empty object
	 * @return {mixed} - Initial data in internal format
	 */
	_initialData() {
		return {};
	}

	getDataTime() {
		return this._data ? this._data._t || 0 : 0;
	}
}

module.exports = ModelBase;