'use strict';
const Util = require('../../Util/Util.js');

class ProviderBase {
	/**
	 * This method is responsible for updating the data associated to the provided id.
	 * @param {string} id - The key of the data to be saved
	 * @param {mixed} data - The actual data, it should be converted in the middleware layer to the proper format
	 * @param {object} options - Custom options for the provider, since it is connection dependent, middlewares should fill it, to make the data consumer implementation independent
	 * @return {Promise} - The promise's resolve will provide and array consisting of [id, data, options]
	 */
	saveData(id, data, options) {
		return Util.createErrorPromise("Method saveData is not implemented in provider.");
	}

	/**
	 * This method is responsible for loading a data associated to the id provided.
	 * @param {string} id - The key of the data to be loaded
	 * @param {object} options - Custom options for the provider, since it is connection dependent, middlewares should fill it, to make the data consumer implementation independent
	 * @return {Promise} - The promise's resolve will provide and array consisting of [id, data, options], where the data is the freshly queried data
	 */
	loadData(id, options) {
		return Util.createErrorPromise("Method loadData is not implemented in provider.");
	}

	/**
	 * This method is responsible for creating non-existing data.
	 * @param {mixed} data - The actual data, it should be converted in the middleware layer to the proper format
	 * @param {object} options - Custom options for the provider, since it is connection dependent, middlewares should fill it, to make the data consumer implementation independent
	 * @return {Promise} - The promise's resolve will provide and array consisting of [id, data, options], the id will be the created key for the record
	 */
	createData(data, options) {
		return Util.createErrorPromise("Method createData is not implemented in provider.");
	}

	/**
	 * This method is responsible for listing, the filter should contain all filter specific data, there is no queryu language for that, so it will depend on the implementation anyway.
	 * @param {mixed} filter - The key of the data to be saved
	 * @param {mixed} data - The actual data, it should be converted in the middleware layer to the proper format
	 * @return {Promise} - The promise's resolve will provide and array consisting of [dataMap, token], where the dataMap is an id -> data map of the items
	 * 						the token is a continuation token, null if lising has finished
	 */
	listData(filter, options) {
		return Util.createErrorPromise("Method listData is not implemented in provider.");
	}

	/**
	 * This method is responsible for continue a multi-part listing, the token provided should be the token received by the listData's response
	 * @param {mixed} token - The continuation token which will help the provider to continue the listing
	 * @return {Promise} - The promise's resolve will provide and array consisting of [dataMap, token], where the dataMap is an id -> data map of the items
	 * 						the token is a continuation token, null if lising has finished
	 */
	continueList(token) {
		return Util.createErrorPromise("Method continueList is not implemented in provider.");
	}
}

module.exports = ProviderBase;