'use strict';
const ProviderBase   = require("./ProviderBase.js");
const Util           = require("../../Util/Util.js");

class RESTProvider extends ProviderBase {
	constructor(httpClient, to) {
		super();
		this._client  = httpClient;
		this._timeout = to || 10000;
	}

	_httpRequest(url, method, body, headers) {
		let options = {
			"url"        : url,
			"method"     : method,
			"body"       : body || "",
			"headers"    : headers || {},
			"User-Agent" : "smbfw-agent",
			"timeout"    : this._timeout,
			"jar"        : true,
			resolveWithFullResponse : true
		};

		return this._client(options);
	}

	_getUrl(options, id) {
		let vars = [...(options.pathParts || [])];
		if (id !== undefined && id !== null) vars.push(id);
		return Util.substitutePathPattern(options.pathPattern, vars);
	}

	_handleRequest(id, data, options) {
		return this._httpRequest(
			this._getUrl(options, id),
			options.method,
			data,
			options.headers
		).then(response => [
			id,
			response.body,
			{
				...options,
				response: response
			}
		]).catch(error => {
			try {
				error.error = JSON.parse(error.error);
			} catch (e) {}
			return Promise.reject(error);
		});
	}

	/**
	 * This method is responsible for updating the data associated to the provided id.
	 * @param {string} id - The key of the data to be saved
	 * @param {mixed} data - The actual data, it should be converted in the middleware layer to the proper format
	 * @param {object} options - Custom options for the provider, since it is connection dependent, middlewares should fill it, to make the data consumer implementation independent
	 * @return {Promise} - The promise's resolve will provide and array consisting of [id, data, options]
	 */
	saveData(id, data, options) {
		return this._handleRequest(id, data, options);
	}

	/**
	 * This method is responsible for loading a data associated to the id provided.
	 * @param {string} id - The key of the data to be loaded
	 * @param {object} options - Custom options for the provider, since it is connection dependent, middlewares should fill it, to make the data consumer implementation independent
	 * @return {Promise} - The promise's resolve will provide and array consisting of [id, data, options], where the data is the freshly queried data
	 */
	loadData(id, options) {
		return this._handleRequest(id, undefined, options);
	}

	/**
	 * This method is responsible for creating non-existing data.
	 * @param {mixed} data - The actual data, it should be converted in the middleware layer to the proper format
	 * @param {object} options - Custom options for the provider, since it is connection dependent, middlewares should fill it, to make the data consumer implementation independent
	 * @return {Promise} - The promise's resolve will provide and array consisting of [id, data, options], the id will be the created key for the record
	 */
	createData(data, options) {
		return this._handleRequest(undefined, data, options);
	}

}

module.exports = RESTProvider;
