'use strict';
const ManagerBase          = require("./ManagerBase.js");
const thunk                = require("redux-thunk").default;
const SmbfwActions      = require("../Action/SmbfwActions");
const ActionHandlerManager = require("./ActionHandlerManager");
const Util                 = require("../Util/Util.js");

const { createStore, applyMiddleware } = require("redux");

const {
	InvalidStateException,
	InvalidArgumentException,
	InvalidStateChangeException
} = require('../Exception/Exceptions.js');

class StateManager extends ManagerBase {
	constructor() {
		super();
		this._initialState    = {};
	}

	_init() {
		this._actionHandlers  = {
			"smbfw.businesslogic" : this._generateActionHandler("smbfw.businesslogic", "smbfw.businesslogic", true),
			"smbfw.services"      : this._generateActionHandler("smbfw.services", "smbfw.services", true)
		};
	}

	//this helper method will generate the middleware methods, which will call our _middlewareHandler with the proper arguments
	//it also registeres all supported middleware types
	_generateActionHandler(type, mandatoryPrefix, supressWithPrefix) {
		this._info("Creating action handler:", type);
		return this._smbfw.createSmbfwObject(ActionHandlerManager, type, mandatoryPrefix, supressWithPrefix);
	}

	preStart() {
		this._log("State manager pre-start, creating store");
		this._store = createStore(
			this._reducerHandler.bind(this),
			applyMiddleware(
				this._actionHandlers["smbfw.businesslogic"].generateHandler(),
				this._actionHandlers["smbfw.services"].generateHandler(),
				thunk
			)
		)
	}

	_reducerHandler(state, action) {
		this._debug("Reducible", action);

		var newState = state;
		if (state === undefined) {
			newState = this.initialState();
		}

		if (SmbfwActions.updateState.match(action)) {
			newState = Util.applyStateChangeTree(newState, action.payload.changeTree, action.payload.changeList);
		} else if (SmbfwActions.increaseLock.match(action) || SmbfwActions.decreaseLock.match(action)) {
			newState = {...newState};
			newState.smbfw = {...newState.smbfw} || {};
			newState.smbfw.lock = {...newState.smbfw.lock} || {};

			let locks = newState.smbfw.lock;
			locks[action.payload] = locks[action.payload] !== undefined ? (locks[action.payload] + action.amount) : action.default;
		}

		if (state !== newState) {
			this._debug("State updated from", state, "to", newState)
		} else {
			this._debug("No state change happened");
		}

		return newState;
	}

	getStore() {
		return this._store;
	}

	initialState() {
		var ret = {
			...this._initialState,
			smbfw : {
				smbfw       : this._smbfw,
				currentScreen  : null,
				lastUserAction : Date.now()
			},
			bl : {

			},
			data : {

			}
		}

		this._info("Initial state provided", ret);
		return ret;
	}

	setInitialState(state) {
		if (typeof state !== "object" || state === null) throw new InvalidArgumentException("State must be an object, received '" + (typeof state) + "'");
		this._initialState = state;
	}

	registerBusinessLogicHandler(handler) {
		return this._actionHandlers["smbfw.businesslogic"].addHandler(handler);
	}

	registerServiceHandler(handler) {
		return this._actionHandlers["smbfw.services"].addHandler(handler);
	}

	createFetchMethod(key, def) {
		return () => {
			let store = this.getStore().getState();
			return Util.resolveReference(key, store, def);
		}
	}

	dispatch(...args) {
		this._store.dispatch(...args);
	}
}

StateManager.prototype.PUBLISHED_METHODS = ["getStore", "setInitialState", "createFetchMethod", "dispatch"];
module.exports = StateManager;