'use strict';

const StateManager            = require("./Manager/StateManager.js");
const ScreenManager           = require("./Manager/ScreenManager.js");
const LocalizationManager     = require("./Manager/LocalizationManager.js")
const BusinessLogicManager    = require("./Manager/BusinessLogicManager.js")
const DataAccessManager       = require("./Manager/DataAccessManager.js")
const SmbfwServicesManager = require("./Manager/SmbfwServicesManager.js")
const Logger                  = require('./Util/Logger.js');
const SmbfwActions         = require("./Action/SmbfwActions.js");


const render   = require('react-dom').render;
const Provider = require('react-redux').Provider;
const React    = require('react');
const connect  = require('react-redux').connect;

const {InvalidStateException, InvalidArgumentException} = require('./Exception/Exceptions.js');


class Smbfw {
	constructor(logLevel) {
		Logger.setLogLevel(logLevel || 0);
		Logger.createChannel("smbfw").bindToObject(this);
		this._log("Smbfw created");

		this._started      = false;
		this._stateManager = null;
		this._managers     = [];

		this._initialAction = null;
		this._userActionMonitorInterval = 0;
		this._lastUserAction = Date.now();


		this._info("Creating managers");
		this._createManager("_stateManager", StateManager, "ss");
		this._createManager("_screenManager", ScreenManager, "sm");
		this._createManager("_localizationManager", LocalizationManager, "l");
		this._createManager("_businessLogicManager", BusinessLogicManager, "bl");
		this._createManager("_dataAccessManager", DataAccessManager, "dm");
		this._createManager("_smbfwServicesManager", SmbfwServicesManager, "fsm");

		this._info("Registering default modules");
		this._registerDefaultModules();
	}

	_createManager(field, cls, accessor) {
		this._info("Registering manager:", cls.name);
		this[field]    = this.createSmbfwObject(cls);
		this[accessor] = this[field].generateAccessors();

		this._managers.push(this[field]);
	}

	_registerDefaultModules() {
		this._log("Registering modules");
	}

	_acquireRootElement(rootElement) {
		if (typeof(rootElement) == "string") rootElement = document.getElementById(rootElement);

		if (!rootElement || !rootElement instanceof HTMLElement) {
			throw new InvalidArgumentException("In order to start smbfw, you must provide a valid HTML element or the ID of a DOM element as a string to be the application's root.");
		}

		return rootElement;
	}

	_setupActionHandlerForManagers(registerMethodName, handlerMethodName, logString) {
		var registerMethod = this._stateManager[registerMethodName].bind(this._stateManager);

		this._managers.forEach(manager => {
			let handlerMethod = manager[handlerMethodName];
			if (typeof handlerMethod !== "function") return;

			this._debug(logString, manager);
			registerMethod(handlerMethod.bind(manager));
		});
	}

	setUserActionMonitorInterval(ms) {
		this._userActionMonitorInterval = ms;
	}

	_setupUserActionMonitor() {
		let reported = false;

		let registerUserAction = () => {
			if (reported) return;
			reported = true;
			this._lastUserAction = Date.now();
		}

		document.addEventListener("click", registerUserAction);
		document.addEventListener("touchstart", registerUserAction);
		document.addEventListener("keydown", registerUserAction);

		setInterval(() => {
			if (reported) {
				reported = false;
				this._stateManager.getStore().dispatch(SmbfwActions.updateState({
					"Smbfw.lastUserAction" : this._lastUserAction
				}));
			}
		}, this._userActionMonitorInterval);
	}

	start(rootElement) {
		if (this._started) throw new InvalidStateException("Framework has already been started!");
		this._log("B(l)ooming starts");

		this._info("Calling manager pre-starters");
		this._managers.forEach(manager => manager.preStart());

		var dispatch = this._stateManager.getStore().dispatch.bind(this._stateManager.getStore());
		this._info("Setting dispatchers");
		this._managers.forEach(manager => manager.setDispatch(dispatch));

		this._info("Setting up action handlers");
		this._setupActionHandlerForManagers("registerServiceHandler", "serviceActionHandler", "Registering service action handler for")
		this._setupActionHandlerForManagers("registerBusinessLogicHandler", "businessLogicActionHandler", "Registering business logic action handler for")

		this._info("Calling manager starters");
		this._managers.forEach(manager => manager.start());

		//Validating and getting root element
		this._rootElement   = this._acquireRootElement(rootElement);
		this._started       = true;

		this._info("Rendering application");
		render(
			React.createElement(Provider, {store: this._stateManager.getStore()},
				this._screenManager.getApplication()
			),
			this._rootElement
		)

		if (this._userActionMonitorInterval > 0) {
			this._setupUserActionMonitor();
		}

		if (this._initialAction) {
			this._info("Firing initial action");
			dispatch(this._initialAction());
		}
	}

	createSmbfwObject(cls, ...args) {
		var obj = new cls(...args);
		obj._setSmbfw(this);

		return obj;
	}

	setInitialAction(action) {
		this._initialAction = action;
	}
}

Smbfw.connect = function(mapStateToProps, mapDispatchToProps, mergeProps, options) {
	let newMapStateToProps = (state, ownProps) => {
		let propMap = (typeof mapStateToProps === "function") ? mapStateToProps(state, ownProps) : {};

		propMap.smbfw    = state.smbfw.smbfw;
		propMap.$_l      = state.smbfw.smbfw.l.localize;

		return propMap;
	};

	return connect(newMapStateToProps, mapDispatchToProps, mergeProps, options);
}

module.exports = Smbfw;