'use strict';
const {InvalidArgumentException} = require('../Exception/Exceptions.js');

class Localizer {
	constructor() {
		//raw localization data, first depth is the locale
		this._ldata = {};
		//map localization key -> localized string map for the current locale, built based on the ldata
		this._lmap = {};

		//the current locale in use
		this._locale = "default";
	}

	setLocalizationData(ldata) {
		this._validateLdata(ldata);
		this._ldata = ldata;
		this._buildLocalizationMap();
	}

	setLocale(locale) {
		this._locale = locale;
	}

	getLocale() {
		return this._locale;
	}

	hasLocale(locale) {
		return this._ldata[locale] !== undefined;
	}

	_buildLocalizationMap(ldata, path) {
		if (!ldata) {
			ldata = this._ldata;
			this._lmap =  ldata ? this._buildLocalizationMap(ldata) : {};
			return this._lmap;
		}

		var ret = {};
		for (let i in ldata) {
			let itemPath = (path ? (path + ".") : "") + i;

			if (typeof ldata[i] === "object") {
				var subset = this._buildLocalizationMap(ldata[i], itemPath);
				ret = {...ret, ...subset}
			}

			if (typeof ldata[i] === "string") {
				ret[itemPath] = (ldata[i].indexOf("%s") === -1) ? ldata[i] : ldata[i].split("%s");
			}
		}

		return ret;
	}

	_validateLdata(ldata, path) {
		if (ldata === null || ldata === undefined) {
			throw new InvalidArgumentException("Localization data cannot be empty!");
		}

		for (let i in ldata) {
			let itemPath = (path ? (path + ".") : "") + i;
			let item     = ldata[i];

			if (item === null || item === undefined) {
				throw new InvalidArgumentException("Localization data for key '" + itemPath + "' is not defined!");
			}

			if (Object.getPrototypeOf(item).constructor.name !== "Object" && Object.getPrototypeOf(item).constructor.name !== "String") {
				throw new InvalidArgumentException("Localization data must be object or string but for '" + itemPath + "' it is '" + Object.getPrototypeOf(item).constructor.name + "'!");
			}

			if (Object.getPrototypeOf(item).constructor.name === "Object") {
				this._validateLdata(item, itemPath);
			}
		}
	}

	localize(key, ...params) {
		let localeKey;
		if (key) {
			localeKey = this._locale + "." + key;
		} else {
			localeKey = "missing_localization_key";
		}

		var ret = this._lmap[localeKey] ? this._lmap[localeKey] : (key || localeKey);

		if (typeof ret === "string") return ret;

		var s = "";
		ret.forEach((val, idx, arr) => {
			s += val;
			if (idx == arr.length - 1) return;
			s += params[idx] ? params[idx] : "";
		})

		return s;
	}
}

module.exports = Localizer;